// opal -- online learning with kernel slicing
//  $Id: pa.h 534 2011-03-29 05:15:46Z ynaga $
// Copyright (c) 2009-2011 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef OPAL_PA_H
#define OPAL_PA_H

#include <stdint.h>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <cstring>
#include <cmath>
#include <algorithm>
#include <vector>
#include <functional>
#include <numeric>
#ifdef USE_MT
#include <tr1/random>
#endif
#ifdef USE_HASH
#include <tr1/unordered_map>
#else
#include <map>
#endif
#if ! defined (USE_HASH_TRIE) && ! defined (USE_MAP_TRIE)
#include "dda.h"
#endif

namespace opal {
  // global type alias
  typedef double                fl_t;
  typedef unsigned int          uint;
  typedef unsigned char         uchar;
  typedef unsigned long         ulong;
  typedef std::vector <uint>    fv_t;
  typedef fv_t::const_iterator  fv_it;
  // coefficient for kernel expansion; support s=1 and r=1 case
  // see pecco/polyk.cpp for a general case
  static const int  POLYK_SIZE = 4;
  static const fl_t COEFF[POLYK_SIZE][POLYK_SIZE] = {
    {0,  0,  0,  0},  // 0
    {1,  1,  0,  0},  // 1
    {1,  3,  2,  0},  // 2
    {1,  7, 12,  6},  // 3
  };
  // static variables and functions
  static const size_t KEY_SIZE  = 8;
  static const size_t BUF_SIZE  = 1 << 16;
  static inline bool getLine (FILE* &fp, char* &line, size_t &read) {
#ifdef __APPLE__
    if ((line = fgetln (fp, &read)) == NULL) return false;
#else
    static ssize_t read_ = 0; static size_t size = 0; // static helps inlining
    if ((read_ = getline (&line, &size, fp)) == -1) return false;
    read = read_;
#endif
    *(line + read - 1) = '\0';
    return true;
  }
  static inline void print_err (char const *format, ...) {
    std::va_list argptr;
    va_start (argptr, format);
    vfprintf (stderr, format, argptr);
    va_end   (argptr);
    std::exit (1);
  }
  static inline void _vbEncode (uint fi, uchar* s, uint &len) {
    for (s[len] = (fi & 0x7f); fi >>= 7; s[++len] = (fi & 0x7f))
      s[len] |= 0x80;
    ++len;
  }
  // options
  enum device_t  { Disk, Mem, Null, DEV_UNDEF };
  enum algo_t    { P, PA1, PA2, ALGO_UNDEF };
  enum mode_t    { BOTH, TRAIN, TEST, MODE_UNDEF };
  enum access_t  { RO, RW }; // mode to access data
  static const char device0[][16] = {"Disk", "Mem", "Null", "Disk|Mem|Null"};
  static const char mode0  [][16] = {"BOTH", "TRAIN", "TEST", "BOTH|TRAIN|TEST"};
  static const char algo0  [][16] = {"P", "PA1", "PA2", "P|PA1|PA2"};
  static       char fn0    [][7]  = {"train", "model0", "model", "test"};
#ifdef USE_HASH
  // incremental Folwer / Noll / Vo (FNV) Hash (FNV-1)
  static const uint FNV_PRIME = 0x01000193;
  static const uint FNV_BASIS = 0x811c9dc5;
  struct inc_fnv { // http://isthe.com/chongo/tech/comp/fnv/#FNV-param
    size_t operator() (const uint ret, const uint fi) const
    { return (ret * FNV_PRIME) ^ fi; }
  };
#endif
  // map- or unordered_map-based trie implementation
#if defined (USE_HASH_TRIE) || defined (USE_MAP_TRIE)
  typedef uint64_t edge; // 32bit parend id + 32bit integer label
  struct node {
    uint id;
    int  value;
    node (const uint id_) : id (id_), value (0) {};
  };
#ifdef USE_HASH
  struct hashfun_edge {
    size_t operator () (const edge &e) const
    { return sizeof (size_t) == 8 ? e : inc_fnv () (e >> 32, e & 0xffffffff); }
  };
#endif
#endif
  // feature bag; pack spare feature indices into ordered dense ones
  class fmap {
  private:
    typedef std::vector <std::pair <uint, uint> > counter_t;
    std::vector <uint> _fn2fi;  // replace this with unordered_map if too sparse
    std::vector <uint> _fi2fn;
    counter_t          _counter;
  public:
    fmap () : _fn2fi (), _fi2fn (), _counter () { _fi2fn.push_back (0); }
    // remap features to minimize memory consumption / trie retrieval cost
    void convertFv2Fv (fv_t* fnv) const {
      fv_t::iterator jt = fnv->begin ();
      for (fv_it it = jt; it != fnv->end (); ++it)
        if (uint fi = _fn2fi[*it < _fn2fi.size () ? *it : 0]) { *jt = fi; ++jt; }
      fnv->erase (jt, fnv->end ());
      std::sort (fnv->begin (), jt);
    }
    void convertFv2Fv (fv_t* fnv, uint maxF) { // register if needed
      if (maxF >= _fn2fi.size ()) _fn2fi.resize (maxF + 1, 0); // widen
      fv_t::iterator jt (fnv->begin ());
      for (fv_it it = jt; it != fnv->end (); ++it, ++jt) {
        uint &fi = _fn2fi[*it];
        if (fi == 0) { fi = _fi2fn.size (); _fi2fn.push_back (*it); } // register
        *jt = fi;
      }
      std::sort (fnv->begin (), jt);
    }
    uint fi2fn     (const uint fi) const { return _fi2fn[fi]; }
    void build () { // build feature mapping; empty element should be removed
      if (_fn2fi.size () < _counter.size ()) // widen if needed
        _fn2fi.resize (_counter.size (), 0);
      std::sort (_counter.rbegin (), _counter.rend ()); // or !_N => partial_sort
      for (counter_t::const_iterator it = _counter.begin ();
           it != _counter.end () && it->first; ++it) {
        uint &fi = _fn2fi[it->second];
        if (fi == 0) { fi = _fi2fn.size (); _fi2fn.push_back (it->second); }
      }
      counter_t ().swap (_counter);
    }
    void inc_count (const fv_t *fv, uint maxF) { // sorted features assumed
      for (uint i = static_cast <uint> (_counter.size ()); i <= maxF; ++i)
        _counter.push_back (counter_t::value_type (0, i));
      for (fv_it it = fv->begin (); it != fv->end (); ++it)
        ++_counter[*it].first;
    }
    void clear () { _fn2fi.clear (); _fi2fn.clear (); _counter.clear (); }
  };
  class ex_t {
  private:
    fl_t  _y;    // label or weight
    fv_t* _vec;  // (binary) feature vector
  public:
    ex_t ()                          : _y (0),     _vec (0)       {}
    ex_t (const fl_t y_, fv_t* vec_) : _y (y_),    _vec (vec_)    {}
    ex_t (const ex_t &ex)            : _y (ex._y), _vec (ex._vec) {}
    ex_t& operator= (const ex_t &ex) { _y = ex._y; _vec = ex._vec; return *this;}
    void set (char* ex, const char* const ex_end, fv_t& vec, bool store, fmap* fm = 0) {
      // set label (weight) of feature vectors (support vectors)
      char* err = NULL;
      char* p   = ex;
      const char* ys = ex; while (p != ex_end && *p != ' ') ++p; *p = '\0';
      _y = std::strtof (ys, &err); // label or weight
      // set features
      vec.clear ();
      while (p != ex_end) {
        const char* fs = ++p;
        while (p != ex_end && *p != ':') ++p; *p = '\0'; // fn
        // char* vs = ++p; // assuming binary feature and ignore value
        while (p != ex_end && *p != ' ') ++p; *p = '\0';
        vec.push_back (static_cast <uint> (std::strtol (fs, &err, 10)));
        if (*err != '\0') print_err ("illegal feature index: %s\n", fs);
      }
      _vec = store ? new fv_t (vec) : &vec;
      if (fm) fm->inc_count (_vec, maxF ());
    }
    void set (const fl_t y, fv_t &vec, bool store, fmap* fm = 0) {
      _y = y; _vec = store ? new fv_t (vec) : &vec;
      if (fm) fm->inc_count (_vec, maxF ());
    }
    void print (FILE * fp) const {
      std::fprintf (fp, "%.16g", _y);
      for (fv_it it = _vec->begin (); it != _vec->end (); ++it)
        std::fprintf (fp, " %d:1", *it);
      std::fprintf (fp, "\n");
    }
    // interface
    const fv_it begin    () const { return _vec->begin (); }
    const fv_it end      () const { return _vec->end (); }
    uint        maxF     () const { return _vec->back (); }
    const fl_t& getLabel () const { return _y; }
    fl_t&       getLabel ()       { return _y; }
    const fv_t* getBody  () const { return _vec; }
    fv_t*       getBody  ()       { return _vec; }
    size_t      getSize  () const { return _vec->size (); }
    fl_t        getNorm  () const { return static_cast <fl_t> (_vec->size ()); }
  };
  template <class T>
  class ex_pool {
  public:
    ex_t * init  ()               { return derived ()->init_impl (); }
    void shuffle ()               { derived ()->shuffle_impl (); }
    void put     (const ex_t &x)  { derived ()->put_impl (x); }
    ex_t * get   ()               { return derived ()->get_impl (); }
    void setup   (fmap &fm)       { derived ()->setup_impl (fm); }
    ex_pool () : _vec () {}
    virtual void read (const char* lfn, const size_t M, bool mem, fmap *fm = 0) {
      FILE* reader = std::fopen (lfn, "r");
      if (! reader) print_err ("no such file: %s\n", lfn);
      char buf[BUF_SIZE]; std::setvbuf (reader, &buf[0], _IOFBF, BUF_SIZE);
      char * line = 0;
      size_t read (0), i (0);
      while (getLine (reader, line, read)) {
        if (M && ++i > M) break;
        _x.set (line, line + read - 1, _vec, mem, fm);
        derived ()->put_impl (_x);
      }
      std::fclose (reader);
    }
  protected:
    ex_t _x;
    fv_t _vec;
  private:
    T* derived () { return static_cast <T*> (this); }
    ex_pool (const ex_pool&);
    ex_pool& operator= (const ex_pool&);
  };
  class null_pool : public ex_pool <null_pool> {
  public:
    null_pool  () : ex_pool <null_pool> (), _fp (0), _line (0), _buf (), _read (0), _i (0), _M (0), _fm (0), _s () {}
    ~null_pool () {}
    ex_t * init_impl    () { _i = 0; _line = 0; _read = 0;  std::fseek (_fp, 0, SEEK_SET); return get_impl (); }
    void shuffle_impl   ()
    { std::fprintf (stderr, "WARNING: skip shuffling; try ./rand_shuf.\n"); }
    void read (const char* lfn, const size_t M, bool mem, fmap *fm = 0) {
      _fp = std::fopen (lfn, "r"); _M = M; _fm = fm; // initialize
      if (! _fp) print_err ("no such file: %s\n", lfn);
      std::setvbuf (_fp, &_buf[0], _IOFBF, BUF_SIZE);
      if (_fm) {
        _i = 0;
        while (getLine (_fp, _line, _read)) {
          if (_M && ++_i > _M) break;
          _x.set (_line, _line + _read - 1, _vec, mem, fm);
        }
      }
    }
    void put_impl       (const ex_t) {} // ignore
    ex_t * get_impl () {
      if (_M && ++_i > _M)               return 0;
      if (! getLine (_fp, _line, _read)) return 0;
      _x.set (_line, _line + _read - 1, _vec, false, _fm);
      if (_fm) _fm->convertFv2Fv (_x.getBody ());
      return &_x;
    }
    void setup_impl (fmap &fm) { _fm = &fm; }       // ignore
    FILE * _fp;
    char * _line;
    char   _buf[BUF_SIZE];
    size_t _read;
    size_t _i;
    size_t _M;
    fmap*  _fm;
  private:
    std::vector <uchar> _s;
  };
  class mem_pool : public ex_pool <mem_pool> {
  public:
    typedef std::vector <ex_t>::iterator ex_it;
    mem_pool  () : ex_pool <mem_pool> (), _ex (), _pos (0) {}
    ~mem_pool () {
      for (ex_it it = _ex.begin (); it != _ex.end (); ++it)
        delete it->getBody ();
    }
    ex_t * init_impl  () { _pos = 0; return get_impl (); }
    void shuffle_impl ()
    { std::random_shuffle (_ex.begin (), _ex.end (), _gen); }
    void put_impl (const ex_t &x) { _ex.push_back (x); }
    ex_t * get_impl () { return (_pos == _ex.size ()) ? 0 : &_ex[_pos++]; }
    void setup_impl (fmap &fm) {
      for (ex_it it = _ex.begin (); it != _ex.end (); ++it)
        fm.convertFv2Fv (it->getBody ());
    }
  private:
    std::vector <ex_t> _ex;
    size_t             _pos;
    struct rand_ {
#ifdef USE_MT
      std::tr1::variate_generator <std::tr1::mt19937,
                                   std::tr1::uniform_int <ulong> > gen;
      rand_ () : gen (std::tr1::mt19937 (), std::tr1::uniform_int <ulong> ()) {}
      long operator () (long max) { return gen (max); }
#else
      ulong gen () { // Xorshift RNG; http://www.jstatsoft.org/v08/i14/paper
        static ulong x (123456789), y (362436069), z (521288629), w (88675123);
        ulong t = (x ^ (x << 11)); x = y; y = z; z = w;
        return w = (w ^ (w >> 19)) ^ (t ^ (t >> 8));
      }
      long operator () (long max) { return gen () % max; }
#endif
    } _gen;
  };
  class disk_pool : public ex_pool <disk_pool> {
  public:
    disk_pool  () : ex_pool <disk_pool> (), fp (std::tmpfile ()), _s () {}
    ~disk_pool () { std::fclose (fp); }
    ex_t * init_impl  () { std::fseek (fp, 0, SEEK_SET); return get_impl (); }
    void shuffle_impl ()
    { std::fprintf (stderr, "WARNING: skip shuffling; try ./rand_shuf.\n"); }
    void put_impl (const ex_t &x) {
      if (_s.size () < x.getSize () * KEY_SIZE)
        _s.resize (x.getSize () * KEY_SIZE);
      char y = x.getLabel () > 0 ? +1 : -1;
      uint len (0), prev (0);
      for (fv_it it = x.begin (); it != x.end (); prev = *it, ++it)
        _vbEncode (*it - prev, &_s[0], len);
      std::fwrite (&y,     sizeof (char), 1, fp);
      std::fwrite (&len,   sizeof (uint),  1,   fp);
      std::fwrite (&_s[0], sizeof (uchar), len, fp);
    }
    ex_t * get_impl () {
      char y = 0; if (std::fread (&y, sizeof (char), 1, fp) == 0) return 0;
      uint len = 0; std::fread (&len, sizeof (uint), 1, fp);
      if (_s.size () < len) _s.resize (len);
      std::fread (&_s[0], sizeof (uchar), len, fp);
      _vec.clear ();
      uint r (0), b (0), prev (0);
      for (uint i = 0; i < len; ++i) {
        r += ((_s[i] & 0x7f) << b);
        if (_s[i] & 0x80) { b += 7; } else { _vec.push_back (prev += r); r = b = 0; }
      }
      _x.set (y, _vec, false);
      return &_x;
    }
    void setup_impl (fmap &fm) {
      disk_pool tmp;
      for (ex_t * x = init (); x; x = get ()) {
        fm.convertFv2Fv (x->getBody ());
        tmp.put (*x);
      }
      std::swap (fp, tmp.fp);
    }
    FILE* fp;
  private:
    std::vector <uchar> _s;
  };
  // template <typename Pool>
  class Model {
  private:
    // type alias
    typedef std::vector <uint>    ss_t;
    typedef ss_t::const_iterator  ss_it;
    typedef std::vector <fl_t>    w_t;
    typedef std::vector <ss_t>    f2ss_t;
    //
    static const int  MAX_TRIAL = 8;  // 4 is good enough;
    static const int  BIN_UNIT  = 1;  // 2^# features are stored in one trie
    static const uint MAX_BIN   =
      sizeof (uint) * 8 / BIN_UNIT + (sizeof (uint) * 8 % BIN_UNIT ? 1 : 0);
#ifdef USE_HASH
    struct hashfun_fv { // FNV-1 hash value
      size_t operator() (const fv_t* f) const
      { return std::accumulate (f->begin (), f->end (), FNV_BASIS, inc_fnv ()); }
    };
    struct equal_fv // not optimal; but I don't know why
    { bool operator() (const fv_t* a, const fv_t* b) const { return *a == *b; } };
    typedef std::tr1::unordered_map <fv_t*, uint, hashfun_fv, equal_fv> fv2s_t;
#else
    struct fvpless
    { bool operator() (const fv_t* a, const fv_t* b) const { return *a < *b; } };
    typedef std::map <fv_t*, uint, fvpless> fv2s_t;
#endif
    // variables
    const int          _d;         // degree of polynomial kernel
    const uint         _N;         // limit on # common features
    const algo_t       _algo;      // Perceptron / PA1 / PA2
    const bool         _average;   // averaging parameters
    const bool         _shuffle;   // shuffling data
    const fl_t         _c;         // aggressiveness parameter (PA1 / PA2)
    const size_t       _M;         // limit on # samples
    const device_t     _device;       // event-on-memory / event-on-disk
    fl_t               _bias;      // just keep constant for kernel expansion

    fmap               _fm;        // feature map
    size_t             _nex;       // # processed examples
    uint               _nsv;       // # support vectors
    uint               _nf;        // # features (= max feature id)
    f2ss_t             _f2ss;      // feature id     -> support vector id
    fv2s_t             _fv2s;      // feature vector -> support vector id
    std::vector <bool> _fbit;      // I don't like vector <bool>
    std::vector <fl_t> _alpha;     // support vector id -> (averaged) alpha
    std::vector <ex_t> _sv;        // support vector id -> support vector
    w_t                _w;         // feature id     -> weight
    w_t                _wa;        // feature id     -> (average) weight
    std::vector <fl_t> _polyk;     // poly kernel;        int -> fl_t
    std::vector <fl_t> _spolyk;    // poly kernel sliced; int -> fl_t
#if defined (USE_HASH_TRIE)
    typedef std::tr1::unordered_map <edge, node, hashfun_edge> trie_t;
#elif defined (USE_MAP_TRIE)
    typedef std::map <edge, node> trie_t;
#else
    typedef pda::da <int, -1, -2, MAX_TRIAL, POLYK_SIZE> trie_t;
#endif
    trie_t** _cf_trie;
    w_t*     _cf_w;
#ifdef USE_PRUNE
    struct pn_t { // upper- and lower-bounds of weights
      fl_t pos;
      fl_t neg;
      pn_t () : pos (0), neg (0) {};
      pn_t (const fl_t pos_, const fl_t neg_) : pos (pos_), neg (neg_) {};
      pn_t& operator= (const pn_t &b) { pos = b.pos; neg = b.neg; return *this; }
    };
    std::vector <pn_t> _f2pn;  // feature id -> weight summation of pos/neg svs
    std::vector <pn_t> _bound;
#endif
#ifdef USE_PMT
    struct pm_t {
      uint t;   // time signature
      fl_t pm;  // temporal partial margin
      pm_t () : t (0), pm (0) {};
      pm_t (const uint t_, const fl_t pm_) : t (t_),   pm (pm_)   {};
      pm_t (const pm_t &l_)                : t (l_.t), pm (l_.pm) {};
      pm_t& operator= (const pm_t &l_) { t = l_.t; pm = l_.pm; return *this; }
    };
    template <typename T>
    class ring { // cycling buffer
    public:
      uint end;  // current pos in the ring
      uint t;    // # updates
      ring () : end (0), t (0), _M () {} // _M (1, T ()) {}
      T& operator[] (const size_t index) { return _M[index]; }
      size_t size   () const { return _M.size (); }
      void   resize (uint n) { _M.insert (_M.begin () + end, n - _M.size (), T ()); }
      void   push   (const T &s, bool overt = true) {
        if (! _M.empty () && overt) { // circulate to overwrite
          if (end >= static_cast <uint> (_M.size ())) end = 0;
          _M[end] = s;
        } else {
          _M.insert (_M.begin () + end, s);
        }
        ++end; ++t;
      }
    private:
      std::vector <T> _M;
    };
    std::vector <ring <ex_t> > _f2ss_up;
    trie_t                     _pm_trie;
    std::vector <pm_t>         _pm;
    std::vector <fl_t>         _limk; // condtion for reuse
#endif
    const fl_t _thresh;
    // uncopyable
    Model (const Model&);
    Model& operator= (const Model&);
    // internal functions
    void _precompute_kernel (const size_t nf) {
      for (uint i = static_cast <uint> (_spolyk.size ()); i <= nf; ++i) {
        _spolyk.push_back (std::pow (static_cast <fl_t> (i+1+1), _d)
                           - std::pow (static_cast <fl_t> (i+1), _d));
#ifdef USE_PolyK
        _polyk.push_back  (std::pow (static_cast <fl_t> (i+1), _d));
#endif
      }
    }
    fl_t _getMargin (const fv_it first, const fv_it last) const {
      fl_t m = 0;
      for (fv_it it = first; it != last && *it <= _nf; ++it) m += _w[*it];
      return m;
    }
    fl_t _getMarginPoly (const fv_it first, const fv_it last, const fl_t y = 0) {
      fl_t m = 0;
#ifdef USE_PolyK
      // PKI
      std::vector <uint> dot (_nsv, 0);
      for (fv_it it = first; it != last && *it <= _nf; ++it) {
        const ss_t &ss = _f2ss[*it];
        for (ss_it st = ss.begin (); st != ss.end (); ++st) ++dot[*st];
      }
      for (uint i = 0; i < _nsv; ++i) m += _sv[i].getLabel () * _polyk[dot[i]];
#else
#ifdef USE_PMT
      while (_limk.size () <= static_cast <size_t> (last - first))
        switch (_d) {
          case 1: _limk.push_back (0); break;
          case 2: _limk.push_back (1); break;
          case 3: _limk.push_back (static_cast <fl_t> (_limk.size ()) / 2.0); break;
        }
#endif
      m += _project <RO> (first, last, y);
#endif
      return m;
    }
    // update
    void _addTo (const ex_t &x, const fl_t t) {
      if (x.maxF () > _nf) { // widen
        _nf = x.maxF ();
        _w.resize (_nf + 1, 0);
        if (_average) _wa.resize (_nf + 1, 0);
      }
      for (fv_it it = x.begin (); it != x.end (); ++it) {
        _w[*it] += t;
        if (_average) _wa[*it] += static_cast <fl_t> (_nex) * t;
      }
    }
    void _pushTo (const ex_t &x, const fl_t t) {
      // check support vector overlap; a little overhead
      fv2s_t::const_iterator fit =_fv2s.find (const_cast <fv_t*> (x.getBody ()));
      if (fit == _fv2s.end ()) { // add support vector
        fv_t* p = new fv_t (*x.getBody ());
        _fv2s.insert (fv2s_t::value_type (p, _nsv)).second;
        if (x.maxF () > _nf) {
          _nf = x.maxF ();
          _f2ss.resize (_nf + 1);
#ifdef USE_PRUNE
          _f2pn.resize (_nf + 1);
#endif
#ifdef USE_PMT
          _f2ss_up.resize (_nf + 1);
#endif
          _fbit.resize (_nf + 1, false);
        }
        _precompute_kernel (x.getSize ());
        for (fv_it it = x.begin (); it != x.end (); ++it) {
          _f2ss[*it].push_back (_nsv);
#ifdef USE_PMT
          _f2ss_up[*it].push (ex_t (t, p), !_N || *it <= _N);
#endif
#ifdef USE_PRUNE
          if (t > 0) _f2pn[*it].pos += t; else _f2pn[*it].neg += t;
#endif
        }
        _sv.push_back (ex_t (t, p)); ++_nsv;
        if (_average) _alpha.push_back (static_cast <fl_t> (_nex) * t);
      } else { // just update
        for (fv_it it = x.begin (); it != x.end (); ++it) {
#ifdef USE_PRUNE
          if (t > 0) _f2pn[*it].pos += t; else _f2pn[*it].neg += t;
#endif
#ifdef USE_PMT
          _f2ss_up[*it].push (ex_t (t, const_cast <fv_t*> (fit->first)));
#endif
        }
        _sv[fit->second].getLabel () += t;
        if (_average) _alpha[fit->second] += static_cast <fl_t> (_nex) * t;
      }
#ifndef USE_PolyK
      _project <RW> (x.begin (), x.end (), t); // explicit weights for conjunctive features
#endif
    }
#ifdef USE_PMT
    bool reuse_pm (size_t &pid, const uint i, const uint fi, const uint prev, fl_t &m_, pm_t* &l, bool projected = true) {
      // feature sequence
      ring <ex_t> &ss = _f2ss_up[fi];
#if defined (USE_HASH_TRIE) || defined (USE_MAP_TRIE)
      const edge e ((static_cast <uint64_t> (pid) << 32) | fi);
      trie_t::iterator pit = _pm_trie.insert (std::make_pair (e, node (_pm_trie.size () + 1))).first;
      pid = pit->second.id;
      int& n = pit->second.value;
#else
      char key[KEY_SIZE];
      uint len = 0;
      _vbEncode (fi - prev, reinterpret_cast <uchar *> (&key[0]), len);
      int& n = _pm_trie.update (key, pid, len, 0);
#endif
      if (! n) { // new feature sequence
        _pm.push_back (pm_t ()); n = _pm.size (); l = &_pm[n-1];
        if (projected && _limk[i] > ss.size ())
          ss.resize (static_cast <size_t> (std::ceil (_limk[i])));
        return true;
      }
      l = &_pm[n-1];
      uint nusv = ss.t - l->t; // # newly added support vectors
      if (  projected && nusv > _limk[i])          return true;  // common
      if (! projected && nusv > _f2ss[fi].size ()) return true;  // rare
      m_ += l->pm; // reuse temporal partial margin
      for (uint p = ss.end; nusv > 0; --nusv)  {  // examine newly added
        const ex_t &s = ss[p ? --p : p = ss.size () - 1]; // loop; dirty bit
        uint dot_c = 0;
        for (fv_it sit = s.begin (); *sit <= prev; ++sit)
          dot_c += _fbit[*sit];
        m_ += s.getLabel () * _spolyk[dot_c];
      }
      return false;
    }
#endif
    void _update_weight (int &id, fl_t &t, w_t &fw)
    { if (id) { fw[id-1] += t; } else { fw.push_back (t); id = fw.size (); } }
    template <access_t MODE>
    bool _traverse (size_t &pid, uint fi, fl_t &t, const int d, trie_t* trie, w_t &fw) {
#if defined (USE_HASH_TRIE) || defined (USE_MAP_TRIE)
      trie_t::iterator tit;
      const edge e ((static_cast <uint64_t> (pid) << 32) | fi);
      if (MODE == RO) { // fetch; add up weight to margin
        if ((tit = trie->find (e)) == trie->end ()) return false;
        t += fw[tit->second.value - 1];
      } else {          // store; add/register margin to weight
        tit = trie->insert (std::make_pair (e, node (trie->size () + 1))).first;
        _update_weight (tit->second.value, t, fw);
      }
      pid = tit->second.id; // next node
#else
      char key[KEY_SIZE];
      uint len = 0;
      _vbEncode (fi, reinterpret_cast <uchar *> (&key[0]), len);
      if (MODE == RO) { // fetch; add up weight to margin
        size_t p = 0;
        int n = trie->traverse (key, pid, p, len);
        if (n == trie->NO_PATH)  return false;
        if (n != trie->NO_VALUE) t += fw[n-1];
      } else {          // store; add/register margin to weight
        _update_weight (trie->update (key, pid, len, 0), t, fw);
      }
#endif
      return _d > d;
    }
#ifdef USE_PRUNE
    void _estimate_bound (const fv_it first, fv_it tail) {
      // compute bound from right-to-left
      _bound.resize (tail - first);
      std::vector <pn_t>::iterator p = _bound.end ();
      for (pn_t b = pn_t (0, 0); tail != first; b = *p) {
        *--p = b;
        if (*--tail <= _nf) { // _f2pn.size ()
          const pn_t &fb = _f2pn[*tail];
          const uint cip = std::min (static_cast <uint> (_spolyk.size () - 1),
                                     static_cast <uint> (tail - first));
          p->pos += fb.pos * _spolyk[cip] + fb.neg * _spolyk[0];
          p->neg += fb.neg * _spolyk[cip] + fb.pos * _spolyk[0];
        }
      }
    }
    bool _prune (const fl_t y, const fl_t m, const uint i) const {
      const pn_t &b = _bound[i];
      if (static_cast <int> (y))
        return y * m > _thresh && y * (m + (y > 0 ? b.neg : b.pos)) > _thresh;
      else
        if (m >= 0) return (m + b.neg >= 0); else return (m + b.pos <  0);
    }
#endif
    template <access_t MODE> // explicit projection
    fl_t _project (const fv_it first, const fv_it last, const fl_t y = 0) {
      // polynomial kernel sliced
      fl_t w[POLYK_SIZE], m (0);
      const fv_it end = MODE == RO ? std::upper_bound (first, last, _nf) : last;
      const fv_it cit = _N         ? std::upper_bound (first, end,  _N)  : end;
      if (MODE == RO) {
#ifdef USE_PRUNE
        if (! _spolyk.empty ()) _estimate_bound (first, last);
#endif
        m += _bias;
      } else {
        _bias += COEFF[_d][0] * y;
        for (int i = 1; i < POLYK_SIZE; ++i) w[i-1] = COEFF[_d][i] * y;
      }
#ifdef USE_PMT
      size_t pid  = 0;
      pm_t*  l    = 0;
#endif
      size_t pos[POLYK_SIZE];
      uint prev = 0;
      fv_it it = first;
      for (uint bin = 0; it != cit; ++it) { // dance in the common part;
#ifdef USE_PRUNE
        if (MODE == RO && _prune (y, m, it - first)) { it = end; break; }
#endif
        fl_t m_ = 0;
#ifdef USE_PMT
        if (MODE == RW || reuse_pm (pid, it - first, *it, prev, m_, l)) {
#endif
          for (uint fi = *it >> (bin * BIN_UNIT); fi >>= BIN_UNIT;) ++bin;
          trie_t * trie = _cf_trie[bin];
          w_t     &fw   = _cf_w[bin];
#if defined (USE_HASH_TRIE) || defined (USE_MAP_TRIE)
          size_t* const p (&pos[0]);
#else
          size_t* const p (MODE == RO ? &pos[0] : &(trie->pid[0]));
#endif
          if (_traverse <MODE> (p[0] = 0, *it, MODE == RO ? m_ : w[0], 1, trie, fw))
            for (fv_it jt = first; jt != it; ++jt)
              if (_traverse <MODE> (p[1] = p[0], *jt, MODE == RO ? m_ : w[1], 2, trie, fw))
                for (fv_it kt = first; kt != jt; ++kt)
                  _traverse <MODE> (p[2] = p[1], *kt, MODE == RO ? m_ : w[2], 3, trie, fw);
#ifdef USE_PMT
        }
        if (MODE == RO) *l = pm_t (_f2ss_up[*it].t, m_);
#endif
        if (MODE == RO) { m += m_; prev = *it; _fbit[*it] = true; }
      }
      if (MODE == RO) { // dance in the rare part
        for (; it != end; ++it) {
#ifdef USE_PRUNE
          if (_prune (y, m, it - first)) break;
#endif
          fl_t m_ = 0;
#ifdef USE_PMT
          if (reuse_pm (pid, it - first, *it, prev, m_, l, false)) {
#endif
            const ss_t &ss = _f2ss[*it];
            for (ss_it st = ss.begin (); st != ss.end (); ++st) {
              const ex_t &s    = _sv[*st];
              uint       dot_c = 0;
              for (fv_it sit = s.begin (); *sit <= prev; ++sit)
                dot_c += _fbit[*sit];
              m_ += s.getLabel () * _spolyk[dot_c];
            }
#ifdef USE_PMT
          }
          *l = pm_t (_f2ss_up[*it].t, m_);
#endif
          m += m_; prev = *it; _fbit[*it] = true;
        }
        for (fv_it jt = first; jt != it; ++jt) _fbit[*jt] = false; // reset
      }
      return m;
    }
  public:
    Model (int d = 0, uint N = 0, algo_t algo = PA1, bool average = false, bool shuffle = true, fl_t c = 1.0, size_t M = 0, device_t device = Mem):
      _d (d), _N (N), _algo (algo), _average (average), _shuffle (shuffle), _c (c), _M (M), _device (device), _bias (0), _fm (), _nex (0), _nsv (0), _nf (0), _f2ss (), _fv2s (), _fbit (), _alpha (), _sv (), _w (), _wa (), _polyk (), _spolyk (), _cf_trie (new trie_t*[MAX_BIN]), _cf_w (new w_t[MAX_BIN] ()),
#ifdef USE_PRUNE
      _f2pn (), _bound (),
#endif
#ifdef USE_PMT
      _f2ss_up (), _pm_trie (),_pm (), _limk (),
#endif
      _thresh (_algo == P ? 0 : 1.0) {
      for (uint i = 0; i < MAX_BIN; ++i) _cf_trie[i] = new trie_t;
    }
    ~Model () { // clean up
      for (fv2s_t::iterator it = _fv2s.begin (); it != _fv2s.end (); ++it)
        delete it->first;
      for (uint i = 0; i < MAX_BIN; ++i) delete _cf_trie[i];
      delete [] _cf_trie;
      delete [] _cf_w;
    }
    // clasification interface
    fl_t classify (fv_t &fv) {
      if (_d == 0) {
        return _getMargin (fv.begin (), fv.end ());
      } else {
        _fm.convertFv2Fv (&fv);
        return _getMarginPoly (fv.begin (), fv.end ());
      }
    }
    bool binClassify (fv_t &fv) { return classify (fv) > 0; }
    // training interface
    void set_ex (ex_t &x, const fl_t y, fv_t &vec, bool store, bool flag)
    { x.set (y, vec, store, flag ? &_fm : 0); }
    void process_example (ex_t &x, bool online = true) {
      if (_d > 0 && online)
        _fm.convertFv2Fv (x.getBody (), x.maxF ());
      fl_t y = x.getLabel ();
      fl_t m = (_d == 0) ? _getMargin (x.begin (), x.end ()) :
               _getMarginPoly (x.begin (), x.end (), y);
      m *= y;
      ++_nex;
      if (m <= _thresh) { // m > _thresh
        fl_t &t = y; // delta of weight vector
        switch (_algo) {
          case P:   break; // Perceptron
          case PA1: t *= std::min (_c, (1 - m) / x.getNorm ()); break; // PA-I
          case PA2: t *= (1 - m) / (x.getNorm () + 1.0 / (2.0 * _c)); break; // PA-II
          default: print_err ("unknown learner.");
        }
        if (_d == 0) _addTo (x, t); else _pushTo (x, t);
      }
    }
    void train_from_file (const char* lfn, const uint iter, const char* tfn = 0) {
      switch (_device) {
        case Mem:  train <mem_pool>  (lfn, iter, tfn); break;
        case Disk: train <disk_pool> (lfn, iter, tfn); break;
        case Null: train <null_pool> (lfn, iter, tfn); break;
        default:   break;
      }
    }
    template <typename Pool>
    void train (const char* lfn, const uint iter, const char* tfn = 0) {
      // read examples into pool prior to training
      Pool pool;
      std::fprintf (stderr, "%s examples..",
                    _device == Mem ? "loading" : (_device == Disk ? "caching" : "reading"));
      pool.read (lfn, _M, _device == Mem, _d > 0 ? &_fm : 0);
      std::fprintf (stderr, "done.\n");
      train (pool, iter, tfn); // do
    }
    template <typename Pool>
    void train (Pool &pool, const uint iter, const char* tfn = 0) {
      if (_d > 0) { _fm.build (); pool.setup (_fm); } // feature mapping
      if (_shuffle) pool.shuffle ();                  // shuffling data
      Pool test_pool;
      if (tfn) {
        test_pool.read  (tfn, 0, _device == Mem);
        if (_d > 0) test_pool.setup (_fm);
      }
      // loop
      for (uint i = 1; i <= iter; ++i) {
        for (ex_t * x = pool.init (); x; x = pool.get ())
          process_example (*x, false);
        std::fprintf (stderr, "\r%s%s iter=%-3d #ex=%-8ld ",
                      _average ? "Ave " : "", algo0[_algo], i, _nex);
        if (_d > 0) std::fprintf (stderr, "#SV=%d; ", _nsv);
        if (tfn) Model (_d, _N, _algo, _average).test
                   (test_pool, _w, _wa, _cf_trie, _cf_w, _sv, _alpha, _nex);
      }
    }
    // inherit parameters prior to test
    template <typename Pool>
    void test (Pool &test_pool, const w_t &w, const w_t &wa, trie_t** cf_trie, w_t* cf_w,
               const std::vector <ex_t> &sv, const w_t &alpha, size_t nex) {
      fl_t n = static_cast <fl_t> (nex + 1);
      if (_d == 0) {
        w_t (w).swap (_w); _nf = _w.size () - 1;
        if (_average)
          for (uint fi = 0; fi <= _nf; ++fi) _w[fi] -= wa[fi] / n;
      } else {
        std::swap (_cf_trie, cf_trie);
        for (uint i = 0; i < MAX_BIN; ++i) _cf_w[i].resize (cf_w[i].size ());
        for (uint i = 0; i < sv.size (); ++i)
          _pushTo (sv[i],  sv[i].getLabel () - (_average ? alpha[i] / n : 0));
      }
      test (test_pool);
      if (_d > 0) std::swap (_cf_trie, cf_trie);
    }
    template <typename Pool>
    void test (Pool &pool, const uint debug = 0) {
      uint pp (0), pn (0), np (0), nn (0);
      for (ex_t * x = pool.init (); x; x = pool.get ()) {
        fl_t m = (_d == 0) ? _getMargin (x->begin (), x->end ()) :
                 _getMarginPoly (x->begin (), x->end ());
        if (m >= 0) if (x->getLabel () > 0) ++pp; else ++np;
        else        if (x->getLabel () > 0) ++pn; else ++nn;
        if (debug == 1) std::fprintf (stdout, "%f\n", m);
      }
      std::fprintf (stderr, "acc. %2.3f%% (pp %d) (pn %d) (np %d) (nn %d)\n",
                    (pp + nn) * 100.0 / (pp + pn + np + nn), pp, pn, np, nn);
    }
    void test_on_file (const char* tfn, const uint debug = 0) {
      null_pool pool;
      pool.read (tfn, 0, false, 0);
      if (_d > 0) pool.setup (_fm);
      test (pool, debug);
    }
    bool load (const char * mfn) {
      std::fprintf (stderr, "loading..");
      FILE* reader = std::fopen (mfn, _d == 0 ? "rb" : "r");
      char buf[BUF_SIZE]; std::setvbuf (reader, &buf[0], _IOFBF, BUF_SIZE);
      if (! reader) print_err ("no such file: %s\n", mfn);
      if (_d == 0) {
        if (! reader) print_err ("no such file: %s\n", mfn);
        if (std::fseek (reader, 0, SEEK_END) != 0) return false;
        _nf = static_cast <uint> (std::ftell (reader) / sizeof (fl_t) - 1);
        if (std::fseek (reader, 0, SEEK_SET) != 0) return false;
        _w.resize (_nf + 1);
        if (_average) _wa.resize (_nf + 1, 0);
        if (std::fread (&_w[0], sizeof (fl_t), _nf + 1, reader) < _nf + 1)
          return false;
        size_t nf (0);
        for (uint i = 0; i < _w.size (); ++i)
          if (_w[i] > 0.0 || _w[i] < 0.0) ++nf;
        std::fprintf (stderr, "done (# features = %ld).\n", nf);
      } else {
        ex_t x;
        fv_t vec;
        char * line = 0;
        size_t read = 0;
        mem_pool pool;
        while (getLine (reader, line, read)) {
          if (std::memchr (line, '#', read - 1) != NULL) continue;
          x.set (line, line + read - 1, vec, true, &_fm);
          pool.put (x);
        }
        _fm.build ();
        pool.setup (_fm);
        for (ex_t * x = pool.init (); x; x = pool.get ())
          _pushTo (*x, x->getLabel ());
        size_t ncf (0);
        for (uint i = 0; i < MAX_BIN; ++i) ncf += _cf_w[i].size ();
        std::fprintf (stderr, "done (# features = %ld).\n", ncf);
      }
      std::fclose (reader);
      return true;
    }
    void average (bool reuse = true) {
      fl_t n = static_cast <fl_t> (_nex + 1);
      if (_d == 0)
        for (uint fi = 0; fi <= _nf; ++fi) _w[fi] -= _wa[fi] / n;
      else
        for (uint i = 0; i < _nsv; ++i)
          if (reuse) _pushTo (_sv[i], -_alpha[i] / n);
          else       _sv[i].getLabel () -= _alpha[i] / n;
    }
    void save (const char* fn) { // const
      std::fprintf (stderr, "saving..");
      FILE* fp = std::fopen (fn, _d == 0 ? "wb" : "w");
      char buf[BUF_SIZE]; std::setvbuf (fp, &buf[0], _IOFBF, BUF_SIZE);
      if (_average) average (false); // averaging
      if (_d == 0) {  // linear training
        std::fwrite (&_w[0], sizeof (fl_t), _nf + 1, fp);
      } else  { // kernel; output model in svm_light-compatible format
        // svm_light-compatible header
        std::fprintf (fp, "1 # kernel type\n");
        std::fprintf (fp, "%d # kernel parameter -d\n", _d);
        std::fprintf (fp, "1 # kernel parameter -s\n");
        std::fprintf (fp, "1 # kernel parameter -r\n");
        std::fprintf (fp, "0 # threshold b\n");
        // output support vectors w/ alpha
        for (uint i = 0; i < _nsv; ++i) {
          fv_t * fv = _sv[i].getBody ();
          for (fv_t::iterator it = fv->begin (); it != fv->end (); ++it)
            *it = _fm.fi2fn (*it);
          std::sort (fv->begin (), fv->end ());
          _sv[i].print (fp);
        }
      }
      std::fclose (fp);
      std::fprintf (stderr, "done.\n");
    }
  };
  struct option { // option handler
    char*     com;     // 0.  command name
    char*     train;   // 1.  training data file name
    char*     model0;  // 2.  model file name for initialization
    char*     model;   // 3.  model file name
    char*     test;    // 4.  dev/test data file name
    int       d;       // 5.  polynomial kernel = (sv*x+1)^d; 0 => dot product
    uint      N;       // 6.  maximum number of features expanded (0; unlimited)
    uint      debug;   // 7.  output closed acc. or not
    algo_t    algo;    // 8.  Perceptron / PA1 / PA2
    uint      ave;     // 9.  averaging parameters
    uint      s;       // 10. shuffling training examples
    uint      iter;    // 11. # training iteration
    fl_t      c;       // 12. update aggressiveness
    size_t    M;       // 13. maximum number of example (0; unlimited)
    device_t  device;  // 14. device to maintain examples (default Mem);
    mode_t    mode;

    option (int argc, char** argv) : com (argv[0]), train (fn0[0]), model0 (fn0[1]), model (fn0[2]), test (fn0[3]), d (0), N (0), debug (0), algo (PA1), ave (0), s (1), iter (20), c (1.0), M (0), device (Mem), mode (BOTH) {
      // getOpt
      switch (argc) {
        case 15: device = selectOptValue <device_t> (argv[14], "event=", device0);
        case 14: M      = getOptValue    <size_t>   (argv[13], "M=");
        case 13: c      = getOptValue    <fl_t>     (argv[12], "c=");
        case 12: iter   = getOptValue    <uint>     (argv[11], "iter=");
        case 11: s      = getOptValue    <uint>     (argv[10], "s=");
        case 10: ave    = getOptValue    <uint>     (argv[9],  "ave=");
        case 9:  algo   = selectOptValue <algo_t>   (argv[8],  "a=", algo0);
        case 8:  debug  = getOptValue    <uint>     (argv[7],  "debug=");
        case 7:  N      = getOptValue    <uint>     (argv[6],  "N=");
        case 6:  d      = getOptValue    <int>      (argv[5],  "Kd=");
        default:
          if (argc < 5 || argc > 15) printError ("Wrong # of arguments.");
          if (! iter) printError ("iter must be >= 1");
          if (d > 3)  printError ("d must be <= 3");
          test   = argv[4];
          model  = argv[3];
          model0 = argv[2];
          train  = argv[1];
          setMode  (); // induce appropriate mode
          printOpt ();
      }
    }
    void setMode () {
      if (std::strcmp (train, "-") == 0 && std::strcmp (test, "-") == 0)
        print_err ("specify at least training or test file.\n");
      else if (std::strcmp (test,  "-") == 0) mode = TRAIN;
      else if (std::strcmp (train, "-") == 0) mode = TEST;
      else                                    mode = BOTH;
      if (std::strcmp (model, "-") == 0 && mode != BOTH)
        print_err ("instant mode can be used w/ both train/test files.\n");
      std::fprintf (stderr, "mode: %s\n", mode0[mode]);
    }
    void printOpt () const {
      std::fprintf (stderr,
                    "%s %s %s %s %s [Kd=%d][N=%d][debug=%d][a=%s][ave=%d][s=%d][iter=%d][c=%g][M=%ld][event=%s]\n",
                    com, train, model0, model, test, d, N, debug, algo0[algo],
                    ave, s, iter, c, M, device0[device]);
    }
    void printError (const char* msg, const char* err = "") {
      std::fprintf (stderr, "%s%s\nUsage: ", msg, err);
      algo = ALGO_UNDEF; mode = MODE_UNDEF; device = DEV_UNDEF;
      printOpt ();
      std::exit (1);
    }
    template <typename T>
    T getOptValue (char* opt, const char* prefix) {
      const size_t len = std::strlen (prefix);
      if (std::strncmp (opt, prefix, len) != 0)
        { printError ("Unrecognized option key: ", opt); std::exit (1); }
      char* err;
      T val = static_cast <T> (std::strtod (opt + len, &err));
      if (*err) printError ("Unrecognized option value: ", opt);
      return val;
    }
    template <typename T>
    T selectOptValue (char* opt, const char* prefix, const char op[][16]) {
      const size_t len = std::strlen (prefix);
      if (std::strncmp (opt, prefix, len) != 0)
        printError ("Unrecognized option key: ", opt);
      for (uint i = 0; std::strchr (op[i], '|') == NULL; ++i)
        if (std::strcmp (opt + len, op[i]) == 0) return static_cast <T> (i);
      printError ("Unknown option value: ", opt);
      return static_cast <T> (NULL); // dummy
    }
  };
}
#endif /* OPAL_PA_H */
